package tdd.tp03;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayDeque;
import java.util.HashMap;
import java.util.Vector;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.FileSplit;
import org.apache.hadoop.mapreduce.lib.input.TextInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;


/* Une classe particulière qui contient 
 *  (1) le nom du fichier où se trouve la ligne
 *  (2) la valeur de la deuxième colonne.
 */
class JoinCandidate implements Writable {
	private String fileName = "";
	private int value = 0;
	
	public JoinCandidate () { }
	
	public JoinCandidate(int value, String fileName) {
		this.fileName = fileName;
		this.value = value;
	
	}

	@Override
	public void readFields(DataInput in) throws IOException {
		/* lire les champs depuis un flux d'entrée */
		fileName = in.readUTF();
		value = in.readInt();
		
	}

	@Override
	public void write(DataOutput out) throws IOException {
	/* Écriture des champs sur le flux de sortie.
	 * Attention doit être fait dans le même ordre que la lecture.
	 */
		out.writeUTF(fileName);
		out.writeInt(value);
		
	}
	
	public String getFileName() { return fileName; }
	public int getValue() { return value; }
	
}


class JoinMapper
extends Mapper<Object, Text, IntWritable, JoinCandidate> {

	@Override
	public void map(Object o, Text in, Context ctx) 
			throws IOException, InterruptedException
	{
	
		String s = in.toString();  //On récupère la ligne courante
		String[] ints = s.split("[ ]+");
		if (ints.length == 2) {
			String fileName = ((FileSplit) ctx.getInputSplit()).getPath().getName();

			IntWritable joinKey = new IntWritable(Integer.parseInt(ints[0]));
			JoinCandidate cand = new JoinCandidate(Integer.parseInt(ints[1]), fileName);
			ctx.write(joinKey,cand);
		}
				
	};
	
}

class JoinReducer
extends Reducer<IntWritable, JoinCandidate, Text, NullWritable> {

	
	
	@Override
	public void reduce(IntWritable joinKey, Iterable<JoinCandidate> candidates,
						Context ctx) 
	throws IOException, InterruptedException {
		Configuration conf = ctx.getConfiguration();
		
		HashMap<String,Vector<Integer>> byFile = new HashMap<>();
		for (String fileName : conf.getStringCollection("input_names")) {
			byFile.put(fileName, new Vector<>());
		}
		
		for(JoinCandidate cand : candidates) {
			
			Vector<Integer> elements = byFile.get(cand.getFileName());
			elements.add(cand.getValue());
			byFile.put(cand.getFileName(), elements);
			
		}
		if (byFile.entrySet().stream().allMatch(e -> !e.getValue().isEmpty()))
			crossProduct(new ArrayDeque<>(byFile.values()), "" + joinKey.get(), ctx);
	
	}
	
	private void crossProduct(ArrayDeque<Vector<Integer>> stack, String accu, Context ctx) throws IOException, InterruptedException {
		
		if (stack.isEmpty()) {
			ctx.write(new Text(accu), NullWritable.get());
		} else {
			Vector<Integer> ints = stack.pop(); 
			for (Integer i : ints) {
					crossProduct(stack, accu + ", " + i, ctx);
			}
			stack.push(ints);
		}
	}
	
	
}
	
public class DriverJoin {

	
	
	public static void main(String[] args)
	
	{
		if (args.length < 2) System.exit(2);
		
		try {
			
			//Création d'une nouvelle configuration
			Configuration conf = new Configuration();

			Job job = Job.getInstance(conf, "tdd.tp03");
		    
					
			
			//Le nom de la classe 'main'
			job.setJarByClass(DriverJoin.class);

			//Le format d'entrée du Map
			//TextInputFormat est un alias pour
			//InputFormat<LongWritable, Text>;
			
			job.setInputFormatClass(TextInputFormat.class);
			
			//Initialisation du Map
			job.setMapperClass(JoinMapper.class);
			
			//Initialisation du Reduce
			job.setReducerClass(JoinReducer.class);

			//Sortie du Map
			job.setMapOutputKeyClass(IntWritable.class);
			job.setMapOutputValueClass(JoinCandidate.class);
			
			//Sortie (du Reduce)
			job.setOutputKeyClass(Text.class);
			job.setOutputValueClass(NullWritable.class);
						
			FileOutputFormat.setOutputPath(job, new Path(args[0]));
			
			Vector<String> inputFiles = new Vector<>();
			for(int i = 1; i < args.length; i++) {
				Path p = new Path(args[i]);
				FileInputFormat.addInputPath(job, p);
				inputFiles.add(p.getName());
			}
			job.getConfiguration().setStrings("input_names",inputFiles.toArray(new String[inputFiles.size()]));
			System.err.println(inputFiles);
			System.exit(job.waitForCompletion(true) ?  0 : 1);
			
		} catch (Exception e) {
			System.err.println("Erreur : " + e);
		}
		
	}
	
}
